h3 {
  margin-top: 0px;
}

/* Tables */

table {
  font-family: arial, sans-serif;
  border-collapse: collapse;
  border: 1px solid #bbbbbb;
  width: 100%;
}

td,
tr {
  text-align: left;
  padding: 4px 8px;
}

tr {
  background-color: var(--color-init);
}

tr:nth-child(even) {
  background-color: #ffffff;
}

td:nth-child(even) {
  border-left: 0px;
}

.row:after {
  content: "";
  display: table;
  clear: both;
}

.column {
  float: left;
  width: 50%;
  padding: 10px;
}
