<div class="guide">
  <div>
    <p>
      The QE app allows you to calculate properties in a simple 4-step process:
    </p>
    <ol>
      <li>
        🔍 <strong>Step 1:</strong> Select the structure you want to run.
      </li>
      <li>
        ⚙️ <strong>Step 2:</strong> Select the properties you are interested in.
      </li>
      <li>
        💻 <strong>Step 3:</strong> Choose the computational resources you want to run on and submit your workflow.
      </li>
      <li>
        🚀 <strong>Step 4:</strong> Monitor and view your workflow results.
      </li>
    </ol>
  </div>

  <div>
    <p>
      New users can go straight to the first step and select their structure.
    </p>
    <p>
      Completed workflows can be viewed in the <strong>Calculation history</strong> section.
    </p>
    <p>
      To start a new calculation in a separate tab, click the <strong>New calculation</strong> button.
    </p>
    <p>
      You can also check out the
      <a href="https://aiidalab-qe.readthedocs.io/tutorials/basic.html" target="_blank">basic</a> tutorial to get
      started
      with the Quantum ESPRESSO app, or try out the
      <a href="https://aiidalab-qe.readthedocs.io/tutorials/advanced.html" target="_blank">advanced</a> tutorial to
      learn
      additional features offered by the app.
    </p>
    <p>
      For a more in-depth dive into the app's features, please refer to the
      <a href="https://aiidalab-qe.readthedocs.io/howto/index.html" target="_blank">how-to guides</a>.
    </p>
  </div>

  <div>
    <h3>In-app guides</h3>
    <p>
      Alternatively, you can select one of our in-app guides below to walk through an example use case.
    </p>
  </div>
</div>
