from .result import ElectronicStructureResultsModel, ElectronicStructureResultsPanel

electronic_structure = {
    "result": {
        "panel": ElectronicStructureResultsPanel,
        "model": ElectronicStructureResultsModel,
    },
}
