.info-box {
  display: none;
  margin: 2px;
  padding: 1em;
  border-left: 3px solid #add8e6;
  background-color: #d9edf7;
}
.info-box .widget-html-content {
  line-height: 1.5;
}
.info-box.in-app-guide.show {
  display: flex !important;
}
.guide ol {
  list-style: none;
}
.guide p:last-of-type {
  margin-bottom: 0.5em;
}
.in-app-guide:has(#guide-header),
.in-app-guide:has(#post-guide) {
  background-color: var(--color-success);
  border-color: #8fbc8f;
}
.in-app-guide:has(#guide-header) {
  margin-bottom: 1em;
}
.in-app-guide:has(#post-guide) {
  margin-top: 1em;
}
#post-guide {
  margin: 1em 0;
}
.in-app-guide .alert {
  color: black;
  margin: 10px 0;
  padding: 10px 14px;
  border-width: 0;
  border-left-width: 3px;
}
.in-app-guide .alert.alert-warning {
  border-color: #ffe4c4;
}
.in-app-guide .alert.alert-success {
  border-color: #8fbc8f;
}
.in-app-guide h4 {
  font-weight: bold;
}

.results-panel-guide {
  margin-bottom: 1em;
}
