<div class="about">
  <p>
    The Quantum ESPRESSO app (or QE app for short) is a graphical front end for calculating materials properties using
    <a href="https://www.quantum-espresso.org/" target="_blank">Quantum ESPRESSO</a> (QE).
    Each property is calculated by workflows powered by the <a href="https://www.aiida.net/" target="_bland">AiiDA</a>
    engine, and maintained in the <a href="https://aiida-quantumespresso.readthedocs.io/en/latest/"
      target="_blank">aiida-quantumespresso</a> plugin and many other plugins developed by the AiiDA community.
  </p>
</div>
