.simplified-view {
  flex-direction: row;
  height: 500px;
}

.simplified-view > div:first-child {
  width: 45%;
}

.simplified-view > div:last-child {
  width: 55%;
}

@media (max-width: 1199px) {
  .simplified-view {
    flex-direction: column;
    height: auto;
  }

  .simplified-view > div:first-child,
  .simplified-view > div:last-child {
    width: 100%;
  }
}

.simplified-view .widget-button {
  min-height: 28px;
}

.simplified-view .aiida-node-view-widget {
  flex: 1;
}

.filename-display {
  min-height: 32px;
}

.rolling-output {
  flex: 1;
  min-height: 300px;
}

.rolling-output .widget-html .widget-html-content {
  position: absolute !important;
  width: 100%;
}

.rolling-output .widget-html .widget-html-content pre {
  position: relative !important;
}

/* Process tree */

.simplified-process-tree {
  height: 100%;
  max-height: 500px;
}

.simplified-process-tree .tree-container {
  flex: 1;
  overflow: auto;
}

.simplified-process-tree .tree-container .widget-box {
  overflow: unset;
}

.simplified-process-tree .tree-node-header {
  align-items: center;
  padding-left: 12px;
}

.simplified-process-tree .widget-html-content {
  line-height: 1.5;
}

.simplified-process-tree .tree-node-toggle:focus:enabled,
.simplified-process-tree .tree-node-toggle:hover:enabled,
.simplified-process-tree .tree-node-toggle:active:enabled,
.simplified-process-tree .calculation-link:focus:enabled,
.simplified-process-tree .calculation-link:hover:enabled,
.simplified-process-tree .calculation-link:active:enabled {
  cursor: pointer;
  color: var(--color-info);
  outline: none;
  box-shadow: none;
  background-color: white;
  transition: color 0.15s ease-in-out;
  -webkit-transition: color 0.15s ease-in-out;
  -moz-transition: color 0.15s ease-in-out;
  -ms-transition: color 0.15s ease-in-out;
  -o-transition: color 0.15s ease-in-out;
}

.simplified-process-tree .tree-node-toggle,
.simplified-process-tree .calculation-link {
  width: fit-content;
  line-height: 1.5;
  padding: 0;
  background: white;
}

.simplified-process-tree .tree-node-toggle {
  margin: 0 6px 0 -16px;
}

.simplified-process-tree .tree-node-toggle:hover:enabled,
.simplified-process-tree .tree-node-toggle:active:enabled,
.simplified-process-tree .tree-node-toggle:focus:hover:enabled {
  color: var(--color-info-dark);
}

.simplified-process-tree .tree-node-toggle:focus:enabled {
  color: inherit;
}

.simplified-process-tree .calculation-link {
  color: var(--color-link);
  margin: 0 2px;
}

.simplified-process-tree .calculation-link:hover:enabled,
.simplified-process-tree .calculation-link:active:enabled,
.simplified-process-tree .calculation-link:focus:hover:enabled {
  color: var(--color-info-dark);
}

.simplified-process-tree .calculation-link:active:enabled {
  background-color: var(--color-init);
}

.simplified-process-tree .tree-node-branches {
  height: 0;
  display: none;
}

.simplified-process-tree .tree-node-branches.open {
  height: auto;
  display: block;
}
