<div class="guide">
    <h2>How to use this page</h2>
    <p>
        This page allows you to view, filter, sort, and manage your calculation history.
    </p>
    <h4>Filters and search</h4>
    <ul>
        <li><b>Quick search field:</b> Use the search bar above the table to find jobs by any visible field.</li>
        <li><b>Column-specific filters:</b> Filter data directly in each column using their respective filter options.</li>
        <li><b>Job state dropdown:</b> Filter jobs by their state (e.g., Finished, Running).</li>
        <li><b>Date range picker:</b> Select a start and end date to narrow down jobs based on creation date.</li>
        <li><b>Properties filter:</b> Select one or more properties to narrow the results. Only calculations that include all the selected properties will be displayed. Leave all checkboxes unselected to include calculations regardless of their properties.</li>
    </ul>
    <h4>Table actions</h4>
    <ul>
        <li><b>Editable fields:</b> Edit the <b>label</b> and <b>description</b> of a job directly by clicking on the respective fields.</li>
        <li><b>Inspect:</b> Click on a job's ID link to view detailed information.</li>
        <li><b>Delete:</b> Remove a job by clicking the "Delete" link in its row. A confirmation page will be opened.</li>
        <li><b>Download:</b> Download raw data (i.e. input and output files) and/or the AiiDA archive of a job using the "Download" link.</li>
    </ul>
    <h4>Display options</h4>
    <ul>
        <li><b>Sorting:</b> Click any column header to sort the table by that column.</li>
        <li><b>Column management:</b> Show, hide columns to customize the table view.</li>
        <li><b>Time format:</b> Toggle between "Absolute" (specific dates) and "Relative" (time elapsed).</li>
        <li><b>ID format:</b> Switch between "PK" or "UUID" for job identification.</li>
    </ul>
</div>
