"""Package for the AiiDAlab QE app version.
Plugin (workflow) version will be read from the plugin's __init__.py file.
"""

from aiidalab_qe.version import __version__

__all__ = [
    "__version__",
]
